import datetime
import logging
import resource

from django.db import connection

from djem.utils.table import Table


def _get_mem_mb():
    
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000


def _get_query_count():
    
    return len(connection.queries)


def _get_stat_table_data(monitor, stat):
    
    if stat not in ('time', 'queries', 'mem'):
        raise TypeError('Unknown statistic "{0}".'.format(stat))
    
    total_key = 'total_{0}'.format(stat)
    min_key = 'min_{0}'.format(stat)
    max_key = 'max_{0}'.format(stat)
    avg_key = 'avg_{0}'.format(stat)
    
    # Each statistic should be formatted slightly differently
    stat_format_map = {
        'time': '{0:.4f}',
        'queries': '{0}',
        'mem': '{0:.3f}'
    }
    
    stat_format = stat_format_map[stat]
    
    data = []
    
    def build_data(parent, _indent=0):
        
        parent_total = parent.stats[total_key]
        
        # Sort the monitors by their total historical runtimes
        children = sorted(parent.children.values(), key=lambda m: m.stats[total_key], reverse=True)
        
        for child in children:
            total = child.stats[total_key]
            
            if parent_total:
                pc = total / parent_total * 100
            else:
                pc = 0
            
            data.append((
                '{0}{1}'.format(' ' * _indent, child.name),
                stat_format.format(child.stats[min_key]),
                stat_format.format(child.stats[max_key]),
                stat_format.format(child.stats[avg_key]),
                stat_format.format(total),
                child.stats['count'],
                '{0}{1:.2f}%'.format(' ' * _indent, pc),
            ))
            
            if child.children:
                build_data(child, _indent=_indent + 1)
    
    build_data(monitor)
    
    return data


def _get_stat_table(monitor, title, stats):
    
    # Include a separator between the output for each statistic, using a
    # combination of blank rows and a row to act as a title for the statistic
    include_separator = len(stats) > 1
    stat_heading_map = {
        'time': 'Timing',
        'queries': 'Queries',
        'mem': 'Memory'
    }
    
    footer = 'Totals: {0}'.format(monitor.get_total_string(include_name=False))
    
    t = Table(
        title=title,
        footer=footer
    )
    
    # Add headings manually, as standard row, to avoid doubling the HRs used
    # in the stat titles below
    t.add_row(['Name', 'Minimum', 'Maximum', 'Average', 'Total', 'Count', '%'])
    
    if not include_separator:
        t.add_row(Table.HR)
    
    for stat in stats:
        if include_separator:
            t.add_row(Table.HR)
            t.add_full_width_row(stat_heading_map[stat])
            t.add_row(Table.HR)
        
        t.add_rows(_get_stat_table_data(monitor, stat))
    
    return t.build_table()


class M:
    
    def __init__(self, name, parent=None):
        
        self.name = name
        
        self.parent = parent
        self.children = {}
        self.active_children = 0
        
        if parent:
            parent.children[name] = self
        
        self.stats = None
        
        self.start_time = None
        self.end_time = None
        self.start_mem = None
        self.end_mem = None
        self.start_queries = None
        self.end_queries = None
    
    def _update_stats(self):
        
        if not self.end_time:
            raise Exception('Monitor not started or still running.')
        
        stats = self.stats
        runtime = self.get_seconds()
        mem_usage = self.get_mem_usage()
        query_count = self.get_query_count()
        
        if stats is None:
            self.stats = {
                'count': 1,
                
                'min_time': runtime,
                'max_time': runtime,
                'avg_time': runtime,
                'total_time': runtime,
                
                'min_mem': mem_usage,
                'max_mem': mem_usage,
                'avg_mem': mem_usage,
                'total_mem': mem_usage,
                
                'min_queries': query_count,
                'max_queries': query_count,
                'avg_queries': query_count,
                'total_queries': query_count
            }
            
            return
        
        stats['count'] += 1
        
        stats['total_time'] += runtime
        stats['total_mem'] += mem_usage
        stats['total_queries'] += query_count
        
        stats['avg_time'] = stats['total_time'] / stats['count']
        stats['avg_mem'] = stats['total_mem'] / stats['count']
        stats['avg_queries'] = stats['total_queries'] / stats['count']
        
        if runtime < stats['min_time']:
            stats['min_time'] = runtime
        
        if mem_usage < stats['min_mem']:
            stats['min_mem'] = mem_usage
        
        if query_count < stats['min_queries']:
            stats['min_queries'] = query_count
        
        if runtime > stats['max_time']:
            stats['max_time'] = runtime
        
        if mem_usage > stats['max_mem']:
            stats['max_mem'] = mem_usage
        
        if query_count > stats['max_queries']:
            stats['max_queries'] = query_count
    
    def start(self):
        
        if self.parent:
            self.parent.active_children += 1
        
        self.start_time = datetime.datetime.now()
        self.start_queries = _get_query_count()
        self.start_mem = _get_mem_mb()
    
    def stop(self):
        
        self.end_mem = _get_mem_mb()
        self.end_time = datetime.datetime.now()
        self.end_queries = _get_query_count()
        self._update_stats()
        
        if self.parent:
            self.parent.active_children -= 1
        
        if self.active_children > 0:
            raise Exception('Cannot end a monitor with running children.')
    
    def get_mem_usage(self):
        
        if self.start_mem is None:
            raise Exception('Monitor not started.')
        
        end_mem = self.end_mem
        if not end_mem:
            end_mem = _get_mem_mb()
        
        return end_mem - self.start_mem
    
    def get_query_count(self):
        
        if self.start_queries is None:
            raise Exception('Monitor not started.')
        
        end_queries = self.end_queries
        if not end_queries:
            end_queries = _get_query_count()
        
        return end_queries - self.start_queries
    
    def get_runtime(self):
        
        if self.start_time is None:
            raise Exception('Monitor not started.')
        
        end_time = self.end_time
        if not end_time:
            end_time = datetime.datetime.now()
        
        return end_time - self.start_time
    
    def get_seconds(self):
        
        return self.get_runtime().total_seconds()
    
    def reset(self):
        """
        Reset the start and end times to allow reuse.
        Do not clear statistics.
        """
        
        self.start_mem = None
        self.end_mem = None
        self.start_time = None
        self.end_time = None
    
    def get_total_string(self, include_name=True):
        
        totals = '{0:.4f} seconds, {1} queries, {2:.3f}MB of RAM'.format(
            self.get_seconds(),
            self.get_query_count(),
            self.get_mem_usage()
        )
        
        if include_name:
            totals = '{0}: {1}'.format(self.name, totals)
        
        return totals
    
    def print_mem_stats(self):
        
        if not self.children:
            print('{0}: {1:.3f}MB of RAM'.format(
                self.name,
                self.get_mem_usage()
            ))
            return
        
        print(_get_stat_table(self, 'Memory Usage Results', ('mem',)))
    
    def print_query_stats(self):
        
        if not self.children:
            print('{0}: {1} queries'.format(
                self.name,
                self.get_query_count()
            ))
            return
        
        print(_get_stat_table(self, 'Query Results', ('queries',)))
    
    def print_time_stats(self):
        
        if not self.children:
            print('{0}: {1:.4f} seconds'.format(
                self.name,
                self.get_seconds()
            ))
            return
        
        print(_get_stat_table(self, 'Timing Results', ('time',)))
    
    def print_stats(self):
        
        if not self.children:
            print(self.get_total_string())
            return
        
        print(_get_stat_table(self, 'Monitor Results', ('time', 'queries', 'mem')))
    
    def __str__(self):
        
        if not self.end_time:
            return '{0}: Running'.format(self.name)
        
        return self.get_total_string()


class Mon:
    
    monitors = {}
    
    last_m = None
    
    @classmethod
    def start(cls, name):
        
        if name in cls.monitors:
            print('Warning: Starting a monitor which has not been ended ({0})'.format(name))
        
        # Re-use an existing monitor with the same name on the same parent, if
        # there is one (this allows the same monitor object to build up stats).
        try:
            m = cls.last_m.children[name]
        except (AttributeError, KeyError):
            m = M(name, cls.last_m)
        else:
            m.reset()  # for re-use
        
        cls.monitors[name] = m
        
        # Remember this monitor as the most recently started. It will become
        # the parent of the next monitor that is started.
        cls.last_m = m
        
        m.start()
    
    @classmethod
    def stop(cls, name):
        
        try:
            m = cls.monitors[name]
        except KeyError:
            raise Exception('Attempted to end a monitor that was never started!')
        
        m.stop()
        
        cls.last_m = m.parent
        del cls.monitors[name]
        
        return m
    
    @classmethod
    def start_qlog(cls):
        
        logger = logging.getLogger('django.db.backends')
        handler = logging.StreamHandler()
        
        cls._qlogger_original_level = logger.level
        cls._qlogger_handler = handler
        
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)
    
    @classmethod
    def stop_qlog(cls):
        
        logger = logging.getLogger('django.db.backends')
        
        logger.setLevel(cls._qlogger_original_level)
        logger.addHandler(cls._qlogger_handler)
    
    @classmethod
    def reset(cls):
        
        cls.monitors.clear()
        cls.last_m = None


def mon(name, allow_recursion=False):
    
    def decorator(fn):
        
        def wrapper(*args, **kwargs):
            
            n = name
            
            if allow_recursion:
                i = 1
                while n in Mon.monitors:
                    n = '_'.join((n, str(i)))
                    i += 1
            
            Mon.start(n)
            result = fn(*args, **kwargs)
            m = Mon.stop(n)
            
            # Only print stats if there is not a parent monitor - this monitor's
            # stats will be reported in its output if there is
            if not m.parent:
                m.print_stats()
            
            return result
        
        return wrapper
    
    return decorator



__version__ = '0.8.0'

# Shortcuts to quick-use utils
from djem.utils.inspect import pp  # NOQA
from djem.utils.mon import M, Mon, mon  # NOQA


class Undefined:
    
    def __bool__(self):
        
        return False
    
    def __str__(self):
        
        return '<undefined>'


UNDEFINED = Undefined()



from django.contrib.auth import get_user_model


class Developer:
    """
    Designed for use within the Django shell to aid developers in performing
    regular operations used for testing and debugging, primarily regarding
    altering aspects of the developer's own User.
    """
    
    def __init__(self, **user_lookup_kwargs):
        
        self.user_lookup_kwargs = user_lookup_kwargs
    
    @property
    def user(self):
        
        if not hasattr(self, '_user'):
            self._user = get_user_model().objects.get(**self.user_lookup_kwargs)
        
        return self._user
    
    def be_awesome(self):
        
        user = self.user
        user.is_staff = True
        user.is_superuser = True
        user.save()
    
    def be_lame(self):
        
        user = self.user
        user.is_staff = False
        user.is_superuser = False
        user.save()
    
    def no_super(self):
        
        user = self.user
        user.is_superuser = False
        user.save()
    
    def no_staff(self):
        
        user = self.user
        user.is_staff = False
        user.save()



import pytz
import warnings

from django.apps import apps
from django.test import SimpleTestCase
from django.utils import timezone

from djem import UNDEFINED
from djem.utils.dt import TimeZoneHelper
from djem.utils.tests import setup_test_app


class UndefinedTestCase(SimpleTestCase):
    
    def test_falsey(self):
        
        self.assertFalse(UNDEFINED)
    
    def test_print(self):
        
        self.assertEqual(str(UNDEFINED), '<undefined>')


class TimeZoneHelperTestCase(SimpleTestCase):
    
    def test_init__good_string(self):
        """
        Test constructing a TimeZoneHelper object with a valid timezone string.
        """
        
        tz = 'Australia/Sydney'
        helper = TimeZoneHelper(tz)
        
        self.assertEqual(helper.tz.zone, tz)
    
    def test_init__bad_string(self):
        """
        Test constructing a TimeZoneHelper object with an invalid timezone string.
        """
        
        with self.assertRaises(pytz.UnknownTimeZoneError):
            TimeZoneHelper('fail')
    
    def test_init__UTC(self):
        """
        Test constructing a TimeZoneHelper object with the pytz UTC singleton.
        """
        
        tz = pytz.UTC
        helper = TimeZoneHelper(tz)
        
        self.assertIs(helper.tz, tz)
    
    def test_init__timezone(self):
        """
        Test constructing a TimeZoneHelper object with a valid pytz timezone
        instance.
        """
        
        tz = pytz.timezone('Australia/Sydney')
        helper = TimeZoneHelper(tz)
        
        self.assertIs(helper.tz, tz)
    
    def test_name(self):
        """
        Test the TimeZoneHelper ``name`` property for several different
        timezones.
        """
        
        self.assertEqual(TimeZoneHelper(pytz.UTC).name, 'UTC')
        self.assertEqual(TimeZoneHelper('Australia/Sydney').name, 'Australia/Sydney')
        self.assertEqual(TimeZoneHelper('US/Eastern').name, 'US/Eastern')
    
    def test_now__utc(self):
        """
        Test the TimeZoneHelper ``now`` method, using the UTC timezone.
        NOTE: This test is subject to some inaccuracy if run at the precise
        moment that caused the two compared datetimes to generate on either
        side of a minute interval.
        """
        
        fmt = '%Y-%m-%d %H:%M'
        
        # Generate the two "now" times as close together as possibly to minimise
        # the difference between them. Compare them with minute-resolution as
        # this is enough to detect timezone mismatches and the chance they will
        # generated either side of a minute interval is less than with seconds.
        now = timezone.now()
        helper_now = TimeZoneHelper(pytz.UTC).now()
        
        self.assertEqual(helper_now.strftime(fmt), now.strftime(fmt))
    
    def test_now__local(self):
        """
        Test the TimeZoneHelper ``now`` method, using a local timezone.
        NOTE: This test is subject to some inaccuracy if run at the precise
        moment that caused the two compared datetimes to generate on either
        side of a minute interval.
        """
        
        fmt = '%Y-%m-%d %H:%M'
        
        # Generate the two "now" times as close together as possibly to minimise
        # the difference between them. Compare them with minute-resolution as
        # this is enough to detect timezone mismatches and the chance they will
        # generated either side of a minute interval is less than with seconds.
        now = timezone.now()
        helper_now = TimeZoneHelper('Australia/Sydney').now()
        
        local = pytz.timezone('Australia/Sydney')
        local_now = local.normalize(now.astimezone(local))
        
        self.assertEqual(helper_now.strftime(fmt), local_now.strftime(fmt))
    
    def test_today__utc(self):
        """
        Test the TimeZoneHelper ``today`` method, using the UTC timezone.
        NOTE: This test is subject to some inaccuracy if run at the precise
        moment that caused the two compared datetimes to generate on either
        side of a day interval.
        """
        
        # It would be pretty unlucky for these two datetimes to generate
        # either side of a day interval.
        today = timezone.now().date()
        helper_today = TimeZoneHelper(pytz.UTC).today()
        
        self.assertEqual(helper_today, today)
    
    def test_today__local(self):
        """
        Test the TimeZoneHelper ``today`` method, using a local timezone.
        NOTE: This test is subject to some inaccuracy if run at the precise
        moment that caused the two compared datetimes to generate on either
        side of a day interval.
        """
        
        # It would be pretty unlucky for these two datetimes to generate
        # either side of a day interval.
        now = timezone.now()
        local = pytz.timezone('Australia/Sydney')
        local_today = local.normalize(now.astimezone(local)).date()
        
        helper_today = TimeZoneHelper('Australia/Sydney').today()
        
        self.assertEqual(helper_today, local_today)
    
    def test_stringify(self):
        """
        Test coercing TimeZoneHelper to a string.
        """
        
        self.assertEqual(str(TimeZoneHelper(pytz.UTC)), 'UTC')
        self.assertEqual(str(TimeZoneHelper('Australia/Sydney')), 'Australia/Sydney')
        self.assertEqual(str(TimeZoneHelper('US/Eastern')), 'US/Eastern')


class SetupTestAppTestCase(SimpleTestCase):
    
    # Use a package *outside* the tests package, since djem's tests themselves
    # make use of setup_test_app()
    package = 'djem.utils'
    
    def test_explicit_label(self):
        """
        Test when an explicit app label is provided. The provided label should
        be used directly.
        """
        
        self.assertNotIn('__test_label__', apps.app_configs)
        setup_test_app(self.package, '__test_label__')
        self.assertIn('__test_label__', apps.app_configs)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('__test_label__')
    
    def test_explicit_label__duplicate(self):
        """
        Test when called multiple times with an explicit app label provided.
        The second attempt should have no effect.
        """
        
        n = len(apps.app_configs)
        
        setup_test_app(self.package, '__test_label__')
        self.assertEqual(len(apps.app_configs), n + 1)
        
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')  # ensure all warnings are triggered
            
            setup_test_app(self.package, '__test_label__')
            
            self.assertEqual(len(w), 1)
            self.assertTrue(w[0].message, 'Attempted setup of duplicate test app "__test_label__".')
            self.assertTrue(w[0].category, RuntimeWarning)
        
        self.assertEqual(len(apps.app_configs), n + 1)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('__test_label__')
    
    def test_implicit_label(self):
        """
        Test when no explicit app label is provided. The label of the
        containing app, suffixed with "_tests", should be used.
        """
        
        self.assertNotIn('djem_tests', apps.app_configs)
        setup_test_app(self.package)
        self.assertIn('djem_tests', apps.app_configs)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('djem_tests')
    
    def test_implicit_label__duplicate(self):
        """
        Test when called multiple times with no explicit app label provided.
        The second attempt should have no effect.
        """
        
        n = len(apps.app_configs)
        
        setup_test_app(self.package)
        self.assertEqual(len(apps.app_configs), n + 1)
        
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')  # ensure all warnings are triggered
            
            setup_test_app(self.package)
            
            self.assertEqual(len(w), 1)
            self.assertTrue(w[0].message, 'Attempted setup of duplicate test app "djem_tests".')
            self.assertTrue(w[0].category, RuntimeWarning)
        
        self.assertEqual(len(apps.app_configs), n + 1)
        
        # Uninstall the app again to avoid polluting other tests
        apps.app_configs.pop('djem_tests')



# Date, time and timezone utils

from django.core.exceptions import ValidationError
from django.utils import timezone

# Allow the file to be imported without pytz installed, though it is required
# to use TimeZoneHelper and other timezone-related functionality
try:
    import pytz
except ImportError:  # pragma: no cover
    pytz = None
    PYTZ_AVAILABLE = False
    TIMEZONE_CHOICES = []
else:
    PYTZ_AVAILABLE = True
    TIMEZONE_CHOICES = [(tz, tz) for tz in pytz.common_timezones]


class TimeZoneHelper:
    
    def __init__(self, tz):
        
        if not PYTZ_AVAILABLE:  # pragma: no cover
            raise RuntimeError('TimeZoneHelper requires pytz to be installed.')
        
        if isinstance(tz, str):
            tz = pytz.timezone(tz)
        
        self.tz = tz
    
    @property
    def name(self):
        
        return self.tz.zone
    
    def now(self):
        """
        Return the current datetime in the local timezone.
        """
        
        utc_now = timezone.now()
        return self.tz.normalize(utc_now.astimezone(self.tz))
    
    def today(self):
        """
        Return the current date in the local timezone.
        """
        
        return self.now().date()
    
    def __str__(self):
        
        return self.name
    
    def __repr__(self):
        
        return '<{0}: {1}>'.format(self.__class__.__name__, str(self))


def get_tz_helper(value):
    """
    Return an instance of TimeZoneHelper based on the given value.
    Valid values are:
     - A timezone string (accepted by pytz.timezone())
     - An instance of pytz.tzinfo.BaseTzInfo
     - The pytz.UTC singleton
    
    Return None if the value is None or the empty string.
    Raise ValidationError if the value cannot be converted into a pytz timezone.
    """
    
    if value in (None, ''):
        return None
    
    if isinstance(value, TimeZoneHelper):
        return value
    
    try:
        return TimeZoneHelper(value)
    except pytz.UnknownTimeZoneError:
        raise ValidationError('Invalid timezone "{0}".'.format(value))
